==> osso-statusbar-cpu_0.4.1-1.tar.gz: osso-statusbar-cpu-0.4.1/src/dialog.c <==
#include <gtk/gtktreeview.h>
#include <gtk/gtktreestore.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkcellrendererprogress.h>
#include <gtk/gtktreeselection.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <hildon-widgets/hildon-note.h>
#include "dialog.h"

enum TopResponse {
	TOP_RESPONSE_KILL = 1,
};

enum TopColumns {
	COL_NAME = 0,
	COL_PID,
	COL_DETAIL,
	COL_PER,
	COL_MODE,
	COL_COUNT,
};

#define _(x) x

static gboolean top_dialog_update(gpointer data);
static gboolean my_isnum(const gchar *subject);
static gint cpu_item_comp(gconstpointer lv, gconstpointer rv);
static gint cpu_sort(gconstpointer lv, gconstpointer rv);
static gint mem_sort(gconstpointer lv, gconstpointer rv);
static void free_cpu_item(gpointer item, gpointer data);
static void free_mem_item(gpointer item, gpointer data);
static const gchar *find_field(FILE *fd, const gchar *field);
static void dialog_response(GtkDialog *dialog, gint response, gpointer data);

struct ProcItemCpu {
	guint ucpu;
	guint scpu;
	guint pid;
	gchar *name;
};

struct ProcItemCpuDelta {
	guint cpu;
	guint pid;
	const gchar *name;
};

struct ProcItemMem {
	guint rss;
	guint pid;
	gchar *name;
};

typedef struct _TopDialog {
	GSList *cpu_hogs;
	gint cycles;
	gint idle_cycles;
	GtkTreeStore *model;
	GtkTreeIter iter_cpu;
	GtkTreeIter iter_mem;
} TopDialog;

void top_dialog(void)
{
	TopDialog top;
	GtkWidget *dialog;
	GtkWidget *view;
	GtkCellRenderer *text_renderer;
	GtkCellRenderer *prog_renderer;
	GtkTreeViewColumn *column;
	GtkTreeIter iter;
	guint i;

	gint this_timeout = 5000;

	top.cpu_hogs = NULL;
	top.cycles = 0;
	top.idle_cycles = 0;

	text_renderer = gtk_cell_renderer_text_new();
	prog_renderer = gtk_cell_renderer_progress_new();
/*	g_object_set(prog_renderer, "text", "", NULL); */
	
	dialog = gtk_dialog_new_with_buttons(_("Process list"),
			NULL,
			GTK_DIALOG_MODAL | GTK_DIALOG_NO_SEPARATOR,
			_("Kill process"), TOP_RESPONSE_KILL,
			_("Close"), GTK_RESPONSE_CLOSE,
			NULL);

	top.model = gtk_tree_store_new(COL_COUNT,
			G_TYPE_STRING, G_TYPE_INT,
			G_TYPE_STRING, G_TYPE_INT,
			G_TYPE_INT);

	gtk_tree_store_append(top.model,
			&top.iter_cpu,
			NULL);
	gtk_tree_store_set(top.model,
			&top.iter_cpu,
			COL_NAME, "CPU hogs",
			COL_MODE, GTK_CELL_RENDERER_MODE_INERT,
			-1);

	for (i = 0; i < 3; i++) {
		gtk_tree_store_append(top.model, &iter, &top.iter_cpu);
		gtk_tree_store_set(top.model, &iter,
				COL_MODE, GTK_CELL_RENDERER_MODE_ACTIVATABLE,
				-1);
	}

	gtk_tree_store_append(top.model,
			&top.iter_mem,
			NULL);
	gtk_tree_store_set(top.model,
			&top.iter_mem,
			COL_NAME, "MEM hogs",
			COL_MODE, GTK_CELL_RENDERER_MODE_INERT,
			-1);

	for (i = 0; i < 3; i++) {
		gtk_tree_store_append(top.model, &iter, &top.iter_mem);
	}

	view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(top.model));
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(view), FALSE);

	column = gtk_tree_view_column_new_with_attributes(_("Process name"),
			text_renderer,
			"text", COL_NAME,
			"mode", COL_MODE,
			NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);
	
	column = gtk_tree_view_column_new_with_attributes(_("Resoure usage"),
			text_renderer,
			"text", COL_DETAIL,
			NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

	column = gtk_tree_view_column_new_with_attributes(_("Percentage"),
			prog_renderer,
			"value", COL_PER,
			"mode", COL_MODE,
			NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
			view);

	g_object_weak_ref(G_OBJECT(dialog), (GWeakNotify)g_source_remove,
			GINT_TO_POINTER(g_timeout_add(this_timeout,
					top_dialog_update,
					&top)));

	gtk_tree_view_expand_all(GTK_TREE_VIEW(view));
	g_signal_connect(view, "realize", G_CALLBACK(gtk_tree_view_expand_all),
			NULL);
	g_signal_connect(view, "row-collapsed", G_CALLBACK(gtk_tree_view_expand_all), NULL);
	g_signal_connect(dialog, "response", G_CALLBACK(dialog_response), view);

	top_dialog_update(&top);
	gtk_widget_show_all(dialog);
	gtk_dialog_run(GTK_DIALOG(dialog));
	
	gtk_widget_destroy(dialog);
}

static gboolean top_dialog_update(gpointer data)
{
	TopDialog *top = (TopDialog *)data;
	GSList *new_list = NULL;
	GDir *proc;
	GSList *old_list = top->cpu_hogs;
	GSList *mem_list = NULL;
	FILE *fd;
	gint user, nice, kern, idle, delta;
	gint mem;
	gint mem_used;
	top->cpu_hogs = NULL;

	fd = fopen("/proc/stat", "r");
	fscanf(fd, "cpu %d %d %d %d", &user, &nice, &kern, &idle);
	fclose(fd);
	delta = user + nice + kern + idle - top->cycles;
	top->cycles = user + nice + kern + idle;

	fd = fopen("/proc/meminfo", "r");
	mem = atoi(find_field(fd, "MemTotal"));
	mem_used = mem;
	mem_used -= atoi(find_field(fd, "MemFree"));
	mem_used -= atoi(find_field(fd, "Buffers"));
	mem_used -= atoi(find_field(fd, "Cached"));
	fclose(fd);

	proc = g_dir_open("/proc", 0, NULL);

	if (proc) {
		GtkTreeIter tree_iter;
		const gchar *file;
		while (( file = g_dir_read_name(proc))) {
			gchar *statfile;
			if (!my_isnum(file)) {
				continue;
			}

			statfile = g_build_filename("/proc", file, "stat",
				       	NULL);

			fd = fopen(statfile, "r");
			g_free(statfile);
			if (fd) {
				struct ProcItemCpu *pitem;
				struct ProcItemCpu item = { 0, 0, 0, NULL };
				GSList *iter;
				fscanf(fd,
					       "%d (%a[^)]) %*1c %*d %*d %*d %*d %*d %*d %*d %*d %*d %*d %d %d",
					       &item.pid, &item.name,
					       &item.ucpu, &item.scpu);
				fclose(fd);
				if ((iter = g_slist_find_custom(old_list,
							&item,
							cpu_item_comp))) {
					
					pitem = (struct ProcItemCpu *)iter->data;
					struct ProcItemCpuDelta *new_item =
						g_new0(struct ProcItemCpuDelta,
							       	1);
					new_item->cpu = item.ucpu + item.scpu -
						pitem->ucpu - pitem->scpu;
					new_item->pid = item.pid;
					new_item->name = item.name;
					free(pitem->name);
					*pitem = item;
					new_list = g_slist_insert_sorted(new_list,
							new_item,
							cpu_sort);
					old_list = g_slist_remove_link(old_list,
							iter);
					iter->data = pitem;
					top->cpu_hogs =
						g_slist_concat(top->cpu_hogs,
								iter);
				} else {
					pitem = g_new0(struct ProcItemCpu, 1);
					*pitem = item;
					top->cpu_hogs = g_slist_prepend(
							top->cpu_hogs,
							pitem);
				}
			}
			statfile = g_build_filename("/proc", file, "status", NULL);
			fd = fopen(statfile, "r");
			g_free(statfile);
			if (fd) {
				struct ProcItemMem *pitem =
					g_new0(struct ProcItemMem, 1);
				pitem->name = g_strdup(find_field(fd, "Name"));
				pitem->pid = atoi(find_field(fd, "Pid"));
				pitem->rss = atoi(find_field(fd, "VmRSS"));
				fclose(fd);

				mem_list = g_slist_insert_sorted(mem_list,
						pitem, mem_sort);
			}
		}
		g_slist_foreach(old_list, free_cpu_item, NULL);
		g_slist_free(old_list);
		g_dir_close(proc);

		gtk_tree_store_set(top->model,
				&top->iter_cpu,
				COL_PER, 
				(delta - idle + top->idle_cycles) * 100 / delta,
				-1);
		if (gtk_tree_model_iter_children(
					GTK_TREE_MODEL(top->model),
					&tree_iter,
					&top->iter_cpu)) {
				do {
					if (new_list != NULL) {
						static gchar buffer[128];
						struct ProcItemCpuDelta *item =
							(struct ProcItemCpuDelta *)
							new_list->data;
						buffer[127] = '\0';
						g_snprintf(buffer, 127,
							"%d cycles",
							item->cpu);
						
						gtk_tree_store_set(
							top->model,
							&tree_iter,
							COL_NAME, item->name,
							COL_PID, item->pid,
							COL_DETAIL, buffer,
							COL_PER, item->cpu *100 / delta,
							-1);
						g_free(new_list->data);
						new_list = g_slist_delete_link(
								new_list,
								new_list);
					} else {
						gtk_tree_store_set(
							top->model,
							&tree_iter,
							COL_NAME, "",
							COL_PID, 0,
							COL_DETAIL, "",
							COL_PER, 0,
							-1);
					}
				} while (gtk_tree_model_iter_next(
							GTK_TREE_MODEL(
								top->model
								), &tree_iter));
		}
		gtk_tree_store_set(top->model,
				&top->iter_mem,
				COL_PER, 
				mem_used * 100 / mem,
				-1);
		if (gtk_tree_model_iter_children(
					GTK_TREE_MODEL(top->model),
					&tree_iter,
					&top->iter_mem)) {
				do {
					if (mem_list != NULL) {
						static gchar buffer[128];
						struct ProcItemMem *item =
							(struct ProcItemMem *)
							mem_list->data;
						buffer[127] = '\0';
						g_snprintf(buffer, 127,
							"%d MB",
							item->rss / 1024);
						
						gtk_tree_store_set(
							top->model,
							&tree_iter,
							COL_NAME, item->name,
							COL_PID, item->pid,
							COL_DETAIL, buffer,
							COL_PER, item->rss *100 / mem,
							-1);
						free_mem_item(mem_list->data, NULL);
						mem_list = g_slist_delete_link(
								mem_list,
								mem_list);
					} else {
						gtk_tree_store_set(
							top->model,
							&tree_iter,
							COL_NAME, "",
							COL_PID, 0,
							COL_DETAIL, "",
							COL_PER, 0,
							-1);
					}
				} while (gtk_tree_model_iter_next(
							GTK_TREE_MODEL(
								top->model
								), &tree_iter));
		}
		g_slist_foreach(mem_list, free_mem_item, NULL);
		g_slist_free(mem_list);
		top->idle_cycles = idle;
	}
	return TRUE;
}

static gboolean my_isnum(const gchar *subject)
{
	if (!*subject)
		return FALSE;
	while (*subject) {
		if (*subject < '0' || *subject > '9')
			return FALSE;
		subject++;
	}

	return TRUE;
}

static gint cpu_item_comp(gconstpointer lv, gconstpointer rv)
{
	const struct ProcItemCpu *lval = 
		(const struct ProcItemCpu *)lv;
	const struct ProcItemCpu *rval =
	       	(const struct ProcItemCpu *)rv;

	return lval->pid - rval->pid;
}

static gint cpu_sort(gconstpointer lv, gconstpointer rv)
{
	const struct ProcItemCpuDelta *lval = 
		(const struct ProcItemCpuDelta *)lv;
	const struct ProcItemCpuDelta *rval =
	       	(const struct ProcItemCpuDelta *)rv;

	return rval->cpu - lval->cpu;
}

static gint mem_sort(gconstpointer lv, gconstpointer rv)
{
	const struct ProcItemMem *lval = 
		(const struct ProcItemMem *)lv;
	const struct ProcItemMem *rval =
	       	(const struct ProcItemMem *)rv;

	return rval->rss - lval->rss;
}

static void free_cpu_item(gpointer item, gpointer data)
{
	const struct ProcItemCpu *cpu =
		(struct ProcItemCpu *)item;

	(void)data;

	free(cpu->name);
	g_free(item);
}

static void free_mem_item(gpointer item, gpointer data)
{
	const struct ProcItemMem *mem =
		(struct ProcItemMem *)item;

	(void)data;

	free(mem->name);
	g_free(item);
}

static const gchar *find_field(FILE *fd, const gchar *field)
{
	static gchar buffer[1024];
	gchar *pos_start;
	gchar *pos_end;
	guint len = strlen(field);
	*buffer = '\0';
	while (!feof(fd) && strncmp(buffer, field, len)) {
		fgets(buffer, 1024, fd);
	}
	if (feof(fd)) {
		*buffer = '\0';
		return buffer;
	}
	pos_start = buffer + len;
	while (*pos_start == ':' || *pos_start == ' ' || *pos_start == '\t')
		pos_start++;
	if ((pos_end = strchr(pos_start, ' '))) {
		*pos_end = '\0';
	}
	if ((pos_end = strchr(pos_start, '\n'))) {
		*pos_end = '\0';
	}
	return pos_start;
}

static void dialog_response(GtkDialog *dialog, gint response, gpointer data)
{
	GtkTreeModel *model;
	GtkTreeIter iter;

	if (response == GTK_RESPONSE_CLOSE) {
		return;
	}

	if (gtk_tree_selection_get_selected(
				gtk_tree_view_get_selection(
					GTK_TREE_VIEW(data)),
				&model,
				&iter)) {
		gint pid;
	       	gchar *name;
		gtk_tree_model_get(
				model,
				&iter,
				COL_PID, &pid,
				COL_NAME, &name,
				-1);
		if (pid != 0) {
			gchar *text = g_strdup_printf(
					"Do you really want to kill %s (%d)?",
				       	name, pid);
			GtkWidget *note = hildon_note_new_confirmation(NULL, text);
			g_free(text);
			if (gtk_dialog_run(GTK_DIALOG(note)) == GTK_RESPONSE_OK) {
				kill(pid, SIGTERM);
			}
			gtk_object_destroy(GTK_OBJECT(note));
		}
		if (name != NULL) {
			g_free(name);
		}
	}
	g_signal_stop_emission_by_name(dialog, "response");
}

==> osso-statusbar-cpu_0.5.7.tar.gz: osso-statusbar-cpu-0.5.7/src/dialog.c <==
#include <gtk/gtktreeview.h>
#include <gtk/gtktreestore.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkcellrendererprogress.h>
#include <gtk/gtktreeselection.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <hildon-widgets/hildon-note.h>
#include "dialog.h"

enum TopResponse {
	TOP_RESPONSE_KILL = 1,
};

enum TopColumns {
	COL_NAME = 0,
	COL_PID,
	COL_DETAIL,
	COL_PER,
	COL_MODE,
	COL_COUNT,
};

#define _(x) x

static gboolean top_dialog_update(gpointer data);
static gboolean my_isnum(const gchar *subject);
static gint cpu_item_comp(gconstpointer lv, gconstpointer rv);
static gint cpu_sort(gconstpointer lv, gconstpointer rv);
static gint mem_sort(gconstpointer lv, gconstpointer rv);
static void free_cpu_item(gpointer item, gpointer data);
static void free_mem_item(gpointer item, gpointer data);
static const gchar *find_field(FILE *fd, const gchar *field);
static void dialog_response(GtkDialog *dialog, gint response, gpointer data);

struct ProcItemCpu {
	guint ucpu;
	guint scpu;
	guint pid;
	gchar *name;
};

struct ProcItemCpuDelta {
	guint cpu;
	guint pid;
	const gchar *name;
};

struct ProcItemMem {
	guint rss;
	guint pid;
	gchar *name;
};

typedef struct _TopDialog {
	GSList *cpu_hogs;
	gint cycles;
	gint idle_cycles;
	GtkTreeStore *model;
	GtkTreeIter iter_cpu;
	GtkTreeIter iter_mem;
} TopDialog;

void top_dialog(void)
{
	TopDialog top;
	GtkWidget *dialog;
	GtkWidget *view;
	GtkCellRenderer *text_renderer;
	GtkCellRenderer *prog_renderer;
	GtkTreeViewColumn *column;
	GtkTreeIter iter;
	guint i;

	gint this_timeout = 5000;

	top.cpu_hogs = NULL;
	top.cycles = 0;
	top.idle_cycles = 0;

	text_renderer = gtk_cell_renderer_text_new();
	prog_renderer = gtk_cell_renderer_progress_new();
/*	g_object_set(prog_renderer, "text", "", NULL); */
	
	dialog = gtk_dialog_new_with_buttons(_("Process list"),
			NULL,
			GTK_DIALOG_MODAL | GTK_DIALOG_NO_SEPARATOR,
			_("Kill process"), TOP_RESPONSE_KILL,
			_("Close"), GTK_RESPONSE_CLOSE,
			NULL);

	top.model = gtk_tree_store_new(COL_COUNT,
			G_TYPE_STRING, G_TYPE_INT,
			G_TYPE_STRING, G_TYPE_INT,
			G_TYPE_INT);

	gtk_tree_store_append(top.model,
			&top.iter_cpu,
			NULL);
	gtk_tree_store_set(top.model,
			&top.iter_cpu,
			COL_NAME, "CPU hogs",
			COL_MODE, GTK_CELL_RENDERER_MODE_INERT,
			-1);

	for (i = 0; i < 3; i++) {
		gtk_tree_store_append(top.model, &iter, &top.iter_cpu);
		gtk_tree_store_set(top.model, &iter,
				COL_MODE, GTK_CELL_RENDERER_MODE_ACTIVATABLE,
				-1);
	}

	gtk_tree_store_append(top.model,
			&top.iter_mem,
			NULL);
	gtk_tree_store_set(top.model,
			&top.iter_mem,
			COL_NAME, "MEM hogs",
			COL_MODE, GTK_CELL_RENDERER_MODE_INERT,
			-1);

	for (i = 0; i < 3; i++) {
		gtk_tree_store_append(top.model, &iter, &top.iter_mem);
	}

	view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(top.model));
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(view), FALSE);

	column = gtk_tree_view_column_new_with_attributes(_("Process name"),
			text_renderer,
			"text", COL_NAME,
			"mode", COL_MODE,
			NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);
	
	column = gtk_tree_view_column_new_with_attributes(_("Resoure usage"),
			text_renderer,
			"text", COL_DETAIL,
			NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

	column = gtk_tree_view_column_new_with_attributes(_("Percentage"),
			prog_renderer,
			"value", COL_PER,
			"mode", COL_MODE,
			NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
			view);

	g_object_weak_ref(G_OBJECT(dialog), (GWeakNotify)g_source_remove,
			GINT_TO_POINTER(g_timeout_add(this_timeout,
					top_dialog_update,
					&top)));

	gtk_tree_view_expand_all(GTK_TREE_VIEW(view));
	g_signal_connect(view, "realize", G_CALLBACK(gtk_tree_view_expand_all),
			NULL);
	g_signal_connect(view, "row-collapsed", G_CALLBACK(gtk_tree_view_expand_all), NULL);
	g_signal_connect(dialog, "response", G_CALLBACK(dialog_response), view);

	top_dialog_update(&top);
	gtk_widget_show_all(dialog);
	gtk_dialog_run(GTK_DIALOG(dialog));
	
	gtk_widget_destroy(dialog);
}

static gboolean top_dialog_update(gpointer data)
{
	TopDialog *top = (TopDialog *)data;
	GSList *new_list = NULL;
	GDir *proc;
	GSList *old_list = top->cpu_hogs;
	GSList *mem_list = NULL;
	FILE *fd;
	gint user, nice, kern, idle, delta;
	gint mem;
	gint mem_used;
	top->cpu_hogs = NULL;

	fd = fopen("/proc/stat", "r");
	fscanf(fd, "cpu %d %d %d %d", &user, &nice, &kern, &idle);
	fclose(fd);
	delta = user + nice + kern + idle - top->cycles;
	top->cycles = user + nice + kern + idle;

	fd = fopen("/proc/meminfo", "r");
	mem = atoi(find_field(fd, "MemTotal"));
	mem_used = mem;
	mem_used -= atoi(find_field(fd, "MemFree"));
	mem_used -= atoi(find_field(fd, "Buffers"));
	mem_used -= atoi(find_field(fd, "Cached"));
	fclose(fd);

	proc = g_dir_open("/proc", 0, NULL);

	if (proc) {
		GtkTreeIter tree_iter;
		const gchar *file;
		while (( file = g_dir_read_name(proc))) {
			gchar *statfile;
			gchar *progname;
			if (!my_isnum(file)) {
				continue;
			}

			statfile = g_build_filename("/proc", file, "cmdline",
					NULL);
			g_file_get_contents(statfile, &progname, NULL, NULL);
			g_free(statfile);

			if (progname) {
				if (*progname) {
					gchar *iter = strchr(progname, ' ');
					if (iter) {
						*iter = '\0';
					}
					iter = strrchr(progname, '/');
					if (iter) {
						memmove(progname, iter + 1,
								strlen(iter));
					}
				} else {
					g_free(progname);
					progname = NULL;
				}
			}

			statfile = g_build_filename("/proc", file, "stat",
				       	NULL);

			fd = fopen(statfile, "r");
			g_free(statfile);
			if (fd) {
				struct ProcItemCpu *pitem;
				struct ProcItemCpu item = { 0, 0, 0, NULL };
				GSList *iter;
				if (progname) {
					fscanf(fd,
							"%d (%*[^)]) %*1c %*d %*d %*d %*d %*d %*d %*d %*d %*d %*d %d %d",
							&item.pid,
							&item.ucpu, &item.scpu);
					item.name = g_strdup(progname);
				} else {
					fscanf(fd,
							"%d (%a[^)]) %*1c %*d %*d %*d %*d %*d %*d %*d %*d %*d %*d %d %d",
							&item.pid, &item.name,
							&item.ucpu, &item.scpu);
				}
				fclose(fd);
				if ((iter = g_slist_find_custom(old_list,
							&item,
							cpu_item_comp))) {
					
					pitem = (struct ProcItemCpu *)iter->data;
					struct ProcItemCpuDelta *new_item =
						g_new0(struct ProcItemCpuDelta,
							       	1);
					new_item->cpu = item.ucpu + item.scpu -
						pitem->ucpu - pitem->scpu;
					new_item->pid = item.pid;
					new_item->name = item.name;
					free(pitem->name);
					*pitem = item;
					new_list = g_slist_insert_sorted(new_list,
							new_item,
							cpu_sort);
					old_list = g_slist_remove_link(old_list,
							iter);
					iter->data = pitem;
					top->cpu_hogs =
						g_slist_concat(top->cpu_hogs,
								iter);
				} else {
					pitem = g_new0(struct ProcItemCpu, 1);
					*pitem = item;
					top->cpu_hogs = g_slist_prepend(
							top->cpu_hogs,
							pitem);
				}
			}
			statfile = g_build_filename("/proc", file, "status", NULL);
			fd = fopen(statfile, "r");
			g_free(statfile);
			if (fd) {
				struct ProcItemMem *pitem =
					g_new0(struct ProcItemMem, 1);
				if (progname) {
					pitem->name = g_strdup(progname);
				} else {
					pitem->name = g_strdup(find_field(fd,
								"Name"));
				}

				pitem->pid = atoi(find_field(fd, "Pid"));
				pitem->rss = atoi(find_field(fd, "VmRSS"));
				fclose(fd);

				mem_list = g_slist_insert_sorted(mem_list,
						pitem, mem_sort);
			}
			g_free(progname);
		}
		g_slist_foreach(old_list, free_cpu_item, NULL);
		g_slist_free(old_list);
		g_dir_close(proc);

		gtk_tree_store_set(top->model,
				&top->iter_cpu,
				COL_PER, 
				(delta - idle + top->idle_cycles) * 100 / delta,
				-1);
		if (gtk_tree_model_iter_children(
					GTK_TREE_MODEL(top->model),
					&tree_iter,
					&top->iter_cpu)) {
				do {
					if (new_list != NULL) {
						static gchar buffer[128];
						struct ProcItemCpuDelta *item =
							(struct ProcItemCpuDelta *)
							new_list->data;
						buffer[127] = '\0';
						g_snprintf(buffer, 127,
							"%d cycles",
							item->cpu);
						
						gtk_tree_store_set(
							top->model,
							&tree_iter,
							COL_NAME, item->name,
							COL_PID, item->pid,
							COL_DETAIL, buffer,
							COL_PER, item->cpu *100 / delta,
							-1);
						g_free(new_list->data);
						new_list = g_slist_delete_link(
								new_list,
								new_list);
					} else {
						gtk_tree_store_set(
							top->model,
							&tree_iter,
							COL_NAME, "",
							COL_PID, 0,
							COL_DETAIL, "",
							COL_PER, 0,
							-1);
					}
				} while (gtk_tree_model_iter_next(
							GTK_TREE_MODEL(
								top->model
								), &tree_iter));
		}
		gtk_tree_store_set(top->model,
				&top->iter_mem,
				COL_PER, 
				mem_used * 100 / mem,
				-1);
		if (gtk_tree_model_iter_children(
					GTK_TREE_MODEL(top->model),
					&tree_iter,
					&top->iter_mem)) {
				do {
					if (mem_list != NULL) {
						static gchar buffer[128];
						struct ProcItemMem *item =
							(struct ProcItemMem *)
							mem_list->data;
						buffer[127] = '\0';
						g_snprintf(buffer, 127,
							"%d MB",
							item->rss / 1024);
						
						gtk_tree_store_set(
							top->model,
							&tree_iter,
							COL_NAME, item->name,
							COL_PID, item->pid,
							COL_DETAIL, buffer,
							COL_PER, item->rss *100 / mem,
							-1);
						free_mem_item(mem_list->data, NULL);
						mem_list = g_slist_delete_link(
								mem_list,
								mem_list);
					} else {
						gtk_tree_store_set(
							top->model,
							&tree_iter,
							COL_NAME, "",
							COL_PID, 0,
							COL_DETAIL, "",
							COL_PER, 0,
							-1);
					}
				} while (gtk_tree_model_iter_next(
							GTK_TREE_MODEL(
								top->model
								), &tree_iter));
		}
		g_slist_foreach(mem_list, free_mem_item, NULL);
		g_slist_free(mem_list);
		top->idle_cycles = idle;
	}
	return TRUE;
}

static gboolean my_isnum(const gchar *subject)
{
	if (!*subject)
		return FALSE;
	while (*subject) {
		if (*subject < '0' || *subject > '9')
			return FALSE;
		subject++;
	}

	return TRUE;
}

static gint cpu_item_comp(gconstpointer lv, gconstpointer rv)
{
	const struct ProcItemCpu *lval = 
		(const struct ProcItemCpu *)lv;
	const struct ProcItemCpu *rval =
	       	(const struct ProcItemCpu *)rv;

	return lval->pid - rval->pid;
}

static gint cpu_sort(gconstpointer lv, gconstpointer rv)
{
	const struct ProcItemCpuDelta *lval = 
		(const struct ProcItemCpuDelta *)lv;
	const struct ProcItemCpuDelta *rval =
	       	(const struct ProcItemCpuDelta *)rv;

	return rval->cpu - lval->cpu;
}

static gint mem_sort(gconstpointer lv, gconstpointer rv)
{
	const struct ProcItemMem *lval = 
		(const struct ProcItemMem *)lv;
	const struct ProcItemMem *rval =
	       	(const struct ProcItemMem *)rv;

	return rval->rss - lval->rss;
}

static void free_cpu_item(gpointer item, gpointer data)
{
	const struct ProcItemCpu *cpu =
		(struct ProcItemCpu *)item;

	(void)data;

	free(cpu->name);
	g_free(item);
}

static void free_mem_item(gpointer item, gpointer data)
{
	const struct ProcItemMem *mem =
		(struct ProcItemMem *)item;

	(void)data;

	free(mem->name);
	g_free(item);
}

static const gchar *find_field(FILE *fd, const gchar *field)
{
	static gchar buffer[1024];
	gchar *pos_start;
	gchar *pos_end;
	guint len = strlen(field);
	*buffer = '\0';
	while (!feof(fd) && strncmp(buffer, field, len)) {
		fgets(buffer, 1024, fd);
	}
	if (feof(fd)) {
		*buffer = '\0';
		return buffer;
	}
	pos_start = buffer + len;
	while (*pos_start == ':' || *pos_start == ' ' || *pos_start == '\t')
		pos_start++;
	if ((pos_end = strchr(pos_start, ' '))) {
		*pos_end = '\0';
	}
	if ((pos_end = strchr(pos_start, '\n'))) {
		*pos_end = '\0';
	}
	return pos_start;
}

static void dialog_response(GtkDialog *dialog, gint response, gpointer data)
{
	GtkTreeModel *model;
	GtkTreeIter iter;

	if (response == GTK_RESPONSE_CLOSE) {
		return;
	}

	if (gtk_tree_selection_get_selected(
				gtk_tree_view_get_selection(
					GTK_TREE_VIEW(data)),
				&model,
				&iter)) {
		gint pid;
	       	gchar *name;
		gtk_tree_model_get(
				model,
				&iter,
				COL_PID, &pid,
				COL_NAME, &name,
				-1);
		if (pid != 0) {
			gchar *text = g_strdup_printf(
					"Do you really want to kill %s (%d)?",
				       	name, pid);
			GtkWidget *note = hildon_note_new_confirmation(NULL, text);
			g_free(text);
			if (gtk_dialog_run(GTK_DIALOG(note)) == GTK_RESPONSE_OK) {
				kill(pid, SIGTERM);
			}
			gtk_object_destroy(GTK_OBJECT(note));
		}
		if (name != NULL) {
			g_free(name);
		}
	}
	g_signal_stop_emission_by_name(dialog, "response");
}
